@@ -1,14 +1,13 @@
 # -*- coding: utf-8 -*-
 import io
 import logging
+import os
 import re
 import socket
 from hashlib import md5
-from contextlib import contextmanager
 
 import boto3
 import boto3.compat
-import boto3.s3.transfer as trans
 from botocore.exceptions import ClientError, ParamValidationError
 from botocore.client import Config
 
@@ -18,10 +17,14 @@ logger = logging.getLogger(__name__)
 
 logging.getLogger('boto3').setLevel(logging.WARNING)
 logging.getLogger('botocore').setLevel(logging.WARNING)
-S3_RETRYABLE_ERRORS = (
-    socket.timeout,
-    trans.ReadTimeoutError, trans.IncompleteReadError
-)
+
+try:
+    from boto3.s3.transfer import S3_RETRYABLE_ERRORS
+except ImportError:
+    S3_RETRYABLE_ERRORS = (
+        socket.timeout,
+    )
+
 try:
     FileNotFoundError
 except NameError:
@@ -61,6 +64,12 @@ def split_path(path):
     else:
         return path.split('/', 1)
 
+key_acls = {'private', 'public-read', 'public-read-write',
+             'authenticated-read', 'aws-exec-read', 'bucket-owner-read',
+             'bucket-owner-full-control'}
+buck_acls = {'private', 'public-read', 'public-read-write',
+             'authenticated-read'}
+
 
 class S3FileSystem(object):
     """
@@ -137,7 +146,10 @@ class S3FileSystem(object):
         """
         anon, key, secret, kwargs, token, ssl = (self.anon, self.key,
                             self.secret, self.kwargs, self.token, self.use_ssl)
-        tok = tokenize(anon, key, secret, kwargs, token, ssl)
+
+        # Include the current PID in the connection key so that different
+        # SSL connections are made for each process.
+        tok = tokenize(anon, key, secret, kwargs, token, ssl, os.getpid())
         if refresh:
             self._conn.pop(tok, None)
         logger.debug("Open S3 connection.  Anonymous: %s", self.anon)
@@ -194,7 +206,7 @@ class S3FileSystem(object):
         self.__dict__.update(state)
         self.s3 = self.connect()
 
-    def open(self, path, mode='rb', block_size=5 * 1024 ** 2):
+    def open(self, path, mode='rb', block_size=5 * 1024 ** 2, acl=''):
         """ Open a file for reading or writing
 
         Parameters
@@ -209,7 +221,7 @@ class S3FileSystem(object):
         if 'b' not in mode:
             raise NotImplementedError("Text mode not supported, use mode='%s'"
                                       " and manage bytes" % (mode[0] + 'b'))
-        return S3File(self, path, mode, block_size=block_size)
+        return S3File(self, path, mode, block_size=block_size, acl=acl)
 
     def _lsdir(self, path, refresh=False):
         if path.startswith('s3://'):
@@ -332,6 +344,28 @@ class S3FileSystem(object):
         """ Return all real keys below path """
         return [f['Key'] for f in self._walk(path, refresh)]
 
+    def chmod(self, path, acl):
+        """ Set Access Control on a bucket/key
+
+        See http://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
+
+        Parameters
+        ----------
+        path : string
+            the object to set
+        acl : string
+            the value of ACL to apply
+        """
+        bucket, key = split_path(path)
+        if key:
+            if acl not in key_acls:
+                raise ValueError('ACL not in %s', key_acls)
+            self.s3.put_object_acl(Bucket=bucket, Key=key, ACL=acl)
+        else:
+            if acl not in buck_acls:
+                raise ValueError('ACL not in %s', buck_acls)
+            self.s3.put_bucket_acl(Bucket=bucket, ACL=acl)
+
     def glob(self, path):
         """
         Find files by glob-matching.
@@ -378,10 +412,10 @@ class S3FileSystem(object):
     def exists(self, path):
         """ Does such a file/directory exist? """
         bucket, key = split_path(path)
-        if key:
+        if key or bucket not in self.ls(''):
             return not raises(FileNotFoundError, lambda: self.ls(path))
         else:
-            return bucket in self.ls('')
+            return True
 
     def cat(self, path):
         """ Returns contents of file """
@@ -402,6 +436,22 @@ class S3FileSystem(object):
         with self.open(path, 'rb', block_size=size) as f:
             return f.read(size)
 
+    def url(self, path, expires=3600):
+        """ Generate presigned URL to access path by HTTP
+
+        Parameters
+        ----------
+        path : string
+            the key path we are interested in
+        expires : int
+            the number of seconds this signature will be good for.
+        """
+        bucket, key = split_path(path)
+        return self.s3.generate_presigned_url(ClientMethod='get_object',
+                                       Params={'Bucket': bucket, 'Key': key},
+                                       ExpiresIn=expires)
+
+
     def get(self, path, filename):
         """ Stream data from file at path to local filename """
         with self.open(path, 'rb') as f:
@@ -422,9 +472,9 @@ class S3FileSystem(object):
                         break
                     f2.write(data)
 
-    def mkdir(self, path):
+    def mkdir(self, path, acl=""):
         """ Make new bucket or empty key """
-        self.touch(path)
+        self.touch(path, acl=acl)
 
     def rmdir(self, path):
         """ Remove empty key or bucket """
@@ -545,7 +595,7 @@ class S3FileSystem(object):
             parent = path.rsplit('/', 1)[0]
             self.dirs.pop(parent, None)
 
-    def touch(self, path):
+    def touch(self, path, acl=""):
         """
         Create empty key
 
@@ -553,11 +603,15 @@ class S3FileSystem(object):
         """
         bucket, key = split_path(path)
         if key:
-            self.s3.put_object(Bucket=bucket, Key=key)
+            if acl and acl not in key_acls:
+                raise ValueError('ACL not in %s', key_acls)
+            self.s3.put_object(Bucket=bucket, Key=key, ACL=acl)
             self.invalidate_cache(path)
         else:
+            if acl and acl not in buck_acls:
+                raise ValueError('ACL not in %s', buck_acls)
             try:
-                self.s3.create_bucket(Bucket=bucket)
+                self.s3.create_bucket(Bucket=bucket, ACL=acl)
                 self.invalidate_cache('')
                 self.invalidate_cache(bucket)
             except (ClientError, ParamValidationError):
@@ -635,10 +689,12 @@ class S3File(object):
     S3FileSystem.open: used to create ``S3File`` objects
     """
 
-    def __init__(self, s3, path, mode='rb', block_size=5 * 2 ** 20):
+    def __init__(self, s3, path, mode='rb', block_size=5 * 2 ** 20, acl=""):
         self.mode = mode
         if mode not in {'rb', 'wb', 'ab'}:
             raise NotImplementedError("File mode must be {'rb', 'wb', 'ab'}, not %s" % mode)
+        if path.startswith('s3://'):
+            path = path[len('s3://'):]
         self.path = path
         bucket, key = split_path(path)
         self.s3 = s3
@@ -652,7 +708,10 @@ class S3File(object):
         self.closed = False
         self.trim = True
         self.mpu = None
+        self.acl = acl
         if mode in {'wb', 'ab'}:
+            if acl and acl not in key_acls:
+                raise ValueError('ACL not in %s', key_acls)
             self.buffer = io.BytesIO()
             self.parts = []
             self.size = 0
@@ -666,12 +725,14 @@ class S3File(object):
                     self.write(s3.cat(path))
                 else:
                     try:
-                        self.mpu = s3.s3.create_multipart_upload(Bucket=bucket, Key=key)
+                        self.mpu = s3.s3.create_multipart_upload(Bucket=bucket,
+                                Key=key, ACL=acl)
                     except (ClientError, ParamValidationError) as e:
                         raise IOError('Open for write failed', path, e)
                     self.loc = self.size
-                    out = self.s3.s3.upload_part_copy(Bucket=self.bucket, Key=self.key,
-                                PartNumber=1, UploadId=self.mpu['UploadId'],
+                    out = self.s3.s3.upload_part_copy(Bucket=self.bucket,
+                                Key=self.key, PartNumber=1,
+                                UploadId=self.mpu['UploadId'],
                                 CopySource=path)
                     self.parts.append({'PartNumber': 1,
                                        'ETag': out['CopyPartResult']['ETag']})
@@ -685,6 +746,11 @@ class S3File(object):
         """ File information about this path """
         return self.s3.info(self.path)
 
+    def url(self):
+        """ HTTP URL to read this file (if it already exists)
+        """
+        return self.s3.url(self.path)
+
     def tell(self):
         """ Current file location """
         return self.loc
@@ -733,7 +799,10 @@ class S3File(object):
             self._fetch(self.start, self.end + self.blocksize)
 
     def __next__(self):
-        return self.readline()
+        out = self.readline()
+        if not out:
+            raise StopIteration
+        return out
 
     next = __next__
 
@@ -845,7 +914,7 @@ class S3File(object):
 
             try:
                 self.mpu = self.mpu or self.s3.s3.create_multipart_upload(
-                                Bucket=self.bucket, Key=self.key)
+                                Bucket=self.bucket, Key=self.key, ACL=self.acl)
             except (ClientError, ParamValidationError) as e:
                 raise IOError('Initating write failed: %s' % self.path, e)
 
@@ -891,7 +960,7 @@ class S3File(object):
                 self.buffer.seek(0)
                 try:
                     self.s3.s3.put_object(Bucket=self.bucket, Key=self.key,
-                                           Body=self.buffer.read())
+                                          Body=self.buffer.read(), ACL=self.acl)
                 except (ClientError, ParamValidationError) as e:
                     raise IOError('Write failed: %s' % self.path, e)
             self.s3.invalidate_cache(self.path)
